# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import base64
import os
import errno
from collections import OrderedDict

def to_byte_string(obj):
    """
    Converts an object into to a Python byte string
    """
    if(isinstance(obj, unicode)):
        return obj.encode("utf-8")
    else:
        return str(obj)

## Previously this class was created using the "namedtuple" function, but
## it didn't work properly. The JSON library that turns Python object
## into JSON string interpreted the binary string as a normal UTF8 string.
class EncapsulatedString:
    def __init__(self, string):
        self.string = string

def encapsulate_strings(python_object):
    """Converts every Python string (_not_ Unicode) to an EncapsulatedString

    Avro DataFileReader yields binary data as a Python string. To be able to jsonify it
    with a custom function we need to convert all strings to an object not understood by
    the JSON encoder. This is a kind of a hack, but on the same time the only solution found
    """
    if isinstance(python_object, str):
        return EncapsulatedString(python_object)
    if isinstance(python_object, list):
        return [encapsulate_strings(e) for e in python_object]
    if isinstance(python_object, OrderedDict):
        new_dict = OrderedDict()
        for (k, v) in python_object.iteritems():
            new_dict[k] = encapsulate_strings(v)
        return new_dict
    if isinstance(python_object, dict):
        return {k: encapsulate_strings(v) for (k, v) in python_object.iteritems()}
    return python_object

class _AvroJSONEncoder(json.JSONEncoder):
    """Custom JSON dumping class for correct handling of Avro bytes fields"""

    def default(self, python_object):
        """Serializes binary data which is otherwise impossible to dump as JSON"""
        if isinstance(python_object, EncapsulatedString):
            return base64.b64encode(python_object.string) 
        else:
            return super(_AvroJSONEncoder, self).default(python_object)

def dict_to_json(python_dict, pretty_print=False):
    """Dumps a Python dictionary to JSON format

    Args:
        python_dict: a Python dictionary containing Avro data_store
    Returns:
        A string being a valid JSON with binary data encoded with Base64
    """
    if pretty_print:
        return _AvroJSONEncoder(indent=4).encode(python_dict)
    else:
        return _AvroJSONEncoder().encode(python_dict)

class FileAlreadyExistsException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from collections import namedtuple

from avroknife.error import error

class EqualitySelection:
    """Specification of a desired value assigned to a key"
    
    The key can be nested, e.g. 'opt1.opt2'.
    """
    def __init__(self, string_):
        """
        Args:
            string_: a string in a form of "opt1.opt2=val"
        """
        key, value = string_.split('=')
        self.__key = key
        self.__value = value
    
    def get_key_parts(self):
        """Returns:
            a list of all nesting components of the key
        """
        return self.__key.split('.')
    
    def get_value(self):
        """Returns:
            value assigned to the key
        """
        return self.__value

class PositionWrtRange(object):
    SMALLER = 1
    INSIDE = 2
    LARGER = 3


class Range:
    """Numerical range"""

    def __init__(self, string):
        """
        Args:
            string: range given in format, e.g., 1-6, -6, 1-. The string
                can be also set to None, which means that range is 
                (-infinity, infinity). """
        self.range_ = [None, None]

        if string is not None:
            parts = string.split("-")
            if len(parts) > 2:
                raise Exception(
                    "Too many elements in range specification '{}'.".format(string))
            elif len(parts) == 1:
                number = int(parts[0])
                self.range_ = [number, number]
            else: ## len(parts) == 2:
                self.range_ = [None, None]
                for i in range(2):
                    if len(parts[i]) != 0:
                        self.range_[i] = int(parts[i])

    def get_position(self, number):
        """@return position of given number with respect to the range"""
        if self.range_[0] is not None:
            if number < self.range_[0]:
                return PositionWrtRange.SMALLER
            else:
                if self.range_[1] is not None:
                    if number > self.range_[1]:
                        return PositionWrtRange.LARGER
                    else:
                        return PositionWrtRange.INSIDE
                else:
                    return PositionWrtRange.INSIDE
        else:
            if self.range_[1] is not None:
                if number <= self.range_[1]:
                    return PositionWrtRange.INSIDE
                else:
                    return PositionWrtRange.LARGER
            else:
                return PositionWrtRange.INSIDE

Record = namedtuple('Record', ['index', 'content'], verbose=False)

class RecordSelector:
    """Allows to iterate over records according to certain selection criteria
    """
    
    def __init__(self, range_=None, selection=None, limit=None):
        """
        Args:
            range_: a Range object. It defines range of accepted record 
                indexes.
            selection: an EqualitySelection object. It defines values of
                accepted properties of records.
            limit: specifies that only this number of records matching all other 
                constraints should be returned.
        """
        if range_ is None:
            range_ = Range(None)
        self.__range = range_
        
        self.__selection = selection
        
        if limit is None:
            limit = sys.maxint
        self.__limit = limit

    @staticmethod
    def __records_in_range(data_store, range_):
        """
        Generates records in the specified range from the given data store
    
        Args:
            data_store: a DataStore object with records
            range_: a Range object specifying the desired records
        Returns:
            records from the given range
        """
        for index, content in enumerate(data_store):
            position = range_.get_position(index)
            if position == PositionWrtRange.SMALLER:
                pass
            elif position == PositionWrtRange.INSIDE:
                yield Record(index, content)
            else: ## position == PositionWrtRange.LARGER:
                raise StopIteration


    @staticmethod
    def __record_fulfills_condition(record, selection):
        """Checks whether a field in Avro records fulfills a condition
    
        Args:
            record: an Avro record as a Python dictionary
            selection: an EqualitySelection object (optional)
        Returns:
            True if a record fulfill the condition
            False otherwise
        """
        if selection is None:
            return True
    
        key_parts = selection.get_key_parts()
        value = selection.get_value()
    
        content = record.content
        try:
            for key_part in key_parts:
                content = content[key_part]
        except KeyError:
            error("Specified key not found in the schema: {}"\
                    .format(".".join(key_parts)))
            raise
    
        if content is None and value=="null":
            return True
        if unicode(content) == unicode(value):
            return True
        else:
            return False
    
    def get_records(self, data_store):
        """
        Args:
            data_store: a DataStore object with records
        Returns:
            Record objects.
        """
        records = self.__records_in_range(data_store, self.__range)
        count = 0
        for record in records:
            if count < self.__limit:
                if self.__record_fulfills_condition(record, self.__selection):
                    count = count+1
                    yield record
            else:
                raise StopIteration



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import string

class Tools:
    @staticmethod
    def replace(text, prefix, replace_function):
        """Replace substring from text marked with a given prefix using
        replace_function. 
        
        Args:
            text: analyzed text
            prefix: prefix that defines a special substring to be replaced
            replace_function: function that accepts found substring and returns
                new value that should replace the substring
        """
        new = []
        while len(text) > 0:
            chunk_index = string.find(text, prefix)
            if chunk_index == -1:
                new.append(text)
                break
            key_start_index = chunk_index + len(prefix)
            key_end_index = string.find(text, ' ', key_start_index)
            if key_end_index == -1:
                key = text[key_start_index:]
            else:
                key = text[key_start_index:key_end_index]
            value = replace_function(key)
            new.append(text[:chunk_index])
            new.append(value)
            if key_end_index == -1:
                text = ''
            else:
                text = text[key_end_index:]
        return ''.join(new)



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import subprocess

from avroknife.test.tools import Tools
from avroknife.test.file_system import HDFS, LocalFS

class RunResult:
    def __init__(self, stdout, input_dir, output_dir):
        self.__stdout = stdout
        self.__input_dir = input_dir
        self.__output_dir = output_dir

    def get_stdout(self):
        """Get produced stdout string"""
        return self.__stdout

    def get_input_path(self, name):
        """Get path to the input in the local file system"""
        return os.path.join(self.__input_dir, name)

    def get_output_path(self, name):
        """Get path to the output in the local file system"""
        return os.path.join(self.__output_dir, name)
    
    def __str__(self):
        return 'stdout: {}\n\ninput_dir={}\noutput_dir={}'.format(
            self.__stdout, self.__input_dir, self.__output_dir)

class CommandLineRunnerException:
    def __init__(self, exception):
        """
        Args:
            exception: instance of subprocess.CalledProcessError
        """
        self.returncode = exception.returncode
        self.cmd = exception.cmd
        self.output = exception.output

class CommandLineRunner:
    __input_file_prefix = '@in:' 
    __output_file_prefix = '@out:'
    __input_subdir = 'input'
    __outputs_subdir = 'outputs'
    __hdfs_path_prefix = ''
    __local_path_prefix = 'local:'

    def __init__(self, program_path, local_input_dir, enforce_local=False):
        """
        Args:
            program_path: path to the command line program to be executed
            local_input_dir: path to the directory containing input files
                that can be referenced through placeholders
            enforce_local: allow running the program only on local file system.
                HDFS is not accessed in this mode.
        """ 
        self.__program_path = program_path
        self.__enforce_local = enforce_local
        self.__local_fs = LocalFS()
        self.__local_tmp_dir = \
            self.__initialize_tmp_dirs(self.__local_fs, local_input_dir)
        if not self.__enforce_local:
            self.__hdfs = HDFS()
            self.__hdfs_tmp_dir = \
                self.__initialize_tmp_dirs(self.__hdfs, local_input_dir)
        self.__is_closed = False

    @staticmethod
    def __initialize_tmp_dirs(fs, local_input_dir):
        dir_ = fs.create_temporary_dir()
        fs.copy_from_local_dir(local_input_dir, 
            fs.join_path([dir_, CommandLineRunner.__input_subdir]))
        fs.create_dir(fs.join_path([dir_, CommandLineRunner.__outputs_subdir]))
        return dir_

    def run(self, args_string, is_input_local, is_output_local, 
            discard_stderr=False):
        """
        Execute program with replacing placeholders in arguments string

        Args:
            args_string: parameters of the program with file placeholders
            is_input_local: if True, the input placeholders will be replaced
                with paths in local file system. If False, they will be replaced
                with paths in HDFS.
            is_output_local: if True, the output placeholders will be replaced
                with path in local file system. If False, they will be replaced
                with paths in HDFS.
            ignore_stderr: if True, the standar output is discarded

        Returns:
            RunResult object

        Raises:
            CommandLineRunnerException: exception raised when executed process
                returns non-zero exit status.
        """
        if self.__is_closed:
            raise Exception('This object has been already closed')
        if self.__enforce_local:
            if not (is_input_local and is_output_local):
                raise Exception('is_input_local={}, is_output_local={}, while '\
                    'the enforce_local mode allows running the program only '\
                    'on local file system '.\
                    format(is_input_local, is_output_local)) 

        local_out_dir = self.__local_fs.create_temporary_dir(
            self.__local_fs.join_path([self.__local_tmp_dir, self.__outputs_subdir]))
        hdfs_out_dir = None
        if not is_output_local:
            hdfs_out_dir = self.__hdfs.create_temporary_dir(
                self.__hdfs.join_path([self.__hdfs_tmp_dir, self.__outputs_subdir]))

        args_replaced = self.__replace_args(args_string, 
            is_input_local, is_output_local, local_out_dir, hdfs_out_dir)

        stdout = self.run_raw(args_replaced, discard_stderr)
        
        if not is_output_local:
            ## We need to delete this directory because the copying operation
            ## requires that the destination directory doesn't already exist
            self.__local_fs.delete_dir(local_out_dir)
            self.__hdfs.copy_to_local_dir(hdfs_out_dir, local_out_dir)

        return RunResult(stdout, 
            os.path.join(self.__local_tmp_dir, self.__input_subdir), 
            local_out_dir)

    def run_raw(self, args_string, discard_stderr=False):
        """
        Execute program WITHOUT replacing placeholders in arguments string

        Args:
            args_string: parameters of the program
            discard_stderr: if True, the standard error is discarded

        Returns:
            stdout string
        """
        return self.__system(self.__program_path + ' ' + args_string, 
            discard_stderr)

    def __replace_args(self, args_string, is_input_local, is_output_local, 
            local_out_dir, hdfs_out_dir):
        text = args_string
        if is_input_local:
            text = self.__replace(text, self.__input_file_prefix,
                self.__local_path_prefix, self.__local_fs.join_path,
                [self.__local_tmp_dir, self.__input_subdir])
        else:
            text = self.__replace(text, self.__input_file_prefix,
                self.__hdfs_path_prefix, self.__hdfs.join_path,
                [self.__hdfs_tmp_dir, self.__input_subdir])
        if is_output_local:
            text = self.__replace(text, self.__output_file_prefix,
                self.__local_path_prefix, self.__local_fs.join_path,
                [local_out_dir])
        else:
            text = self.__replace(text, self.__output_file_prefix,
                self.__hdfs_path_prefix, self.__hdfs.join_path,
                [hdfs_out_dir])
        return text

    @staticmethod
    def __replace(text, placeholder_prefix, path_prefix, path_joiner_function, 
            dir_name_elements):
        """Replace placeholders with paths to files"""

        replaced = Tools.replace(text, placeholder_prefix, 
            lambda s: path_prefix + path_joiner_function(dir_name_elements + [s]))
        return replaced


    def close(self):
        """Do the cleanup"""
        if self.__is_closed:
            raise Exception('This object has been already closed')
        self.__is_closed = True
        self.__local_fs.delete_dir(self.__local_tmp_dir)
        if not self.__enforce_local:
            self.__hdfs.delete_dir(self.__hdfs_tmp_dir)

    @staticmethod
    def __system(command, discard_stderr):
        try:
            if discard_stderr:
                with open(os.devnull, 'w') as devnull:
                    return subprocess.check_output(
                            command, shell=True, stderr=devnull)
            else:
                return subprocess.check_output(command, shell=True)
        except subprocess.CalledProcessError as ex:
            raise CommandLineRunnerException(ex)



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch
import avro
from avro.datafile import DataFileReader
from avro.io import DatumReader, SchemaResolutionException
from collections import OrderedDict

from avroknife.error import error

class _FieldsOrderPreservingDatumReader(DatumReader):
    """DatumReader that preserves the order of the fields as defined in schema. 
    
    This is a "hacked" version of the DatumReader class available in Avro 1.7.5
    in package `io`. The only difference is that the `read_record` variable
    is defined as `OrderedDict` object instead of `dict`.
    """
    def read_record(self, writers_schema, readers_schema, decoder):
        # schema resolution
        readers_fields_dict = readers_schema.fields_dict
        read_record = OrderedDict() ## using OrderedDict instead of dict
        for field in writers_schema.fields:
            readers_field = readers_fields_dict.get(field.name)
            if readers_field is not None:
                field_val = self.read_data(field.type, readers_field.type, decoder)
                read_record[field.name] = field_val
            else:
                self.skip_data(field.type, decoder)
        # fill in default values
        if len(readers_fields_dict) > len(read_record):
            writers_fields_dict = writers_schema.fields_dict
            for field_name, field in readers_fields_dict.items():
                if not writers_fields_dict.has_key(field_name):
                    if field.has_default:
                        field_val = self._read_default_value(field.type, field.default)
                        read_record[field.name] = field_val
                    else:
                        fail_msg = 'No default value for field %s' % field_name
                        raise SchemaResolutionException(fail_msg, writers_schema,
                                                      readers_schema)
        return read_record

class DataStore:
    """Avro data store.
    
    Avro data store is a directory with many Avro files where each one has the
    same schema. All these files are treated as if they were a single,
    concatenated Avro file. They are virtually "concatenated" along with the
    increasing order of the paths to these files.
    """

    def __init__(self, datastore_path, schema_path=None):
        """
        Args:
            datastore_path: a FileSystemPath object. Path to a directory 
                containing Avro files, all of them need to have the same schema.
            schema_path: a FileSystemPath object. Path to file containing
                JSON Avro reader schema.
        """
        self._datastore_path = datastore_path
        self._schema_path = schema_path
        self._schema = None


    def get_schema(self):
        """Lazy accessor for data store schema

        If schema is given as a run parameter, then returns this schema.
        Otherwise extracts the schema from the Avro data store files.
        """
        if self._schema:
            return self._schema
        else:
            if not self._schema_path: #if there is no schema
                paths = self.__get_paths_to_avro_files()
                with DataFileReader(paths[0].open("r"), _FieldsOrderPreservingDatumReader()) as reader:
                    self._schema = avro.schema.parse(reader.get_meta('avro.schema'))
                    return self._schema
            else: #a schema is given
                try:
                    self._schema = avro.schema.parse(self._schema_path.open("r").read())
                    return self._schema
                except TypeError:
                    error("supplied schema cannot be parsed!")
                    raise

    def __iter__(self):
        paths = self.__get_paths_to_avro_files()
        prev_global_record_index = 0
        for path in paths:
            with DataFileReader(path.open("r"), _FieldsOrderPreservingDatumReader(
                    readers_schema=self.get_schema())) as reader:
                prev_local_record_index = 0
                try:
                    for record in reader:
                        yield record
                        prev_global_record_index = prev_global_record_index + 1
                        prev_local_record_index = prev_local_record_index + 1
                except Exception:
                    error("processing record with index {} failed. "\
                        "This record comes from \"{}\" Avro file and in this "\
                        "file it has local index equal {}.".format(
                            prev_global_record_index+1, path,
                            prev_local_record_index+1))
                    raise

    def __get_paths_to_avro_files(self):
        paths = []
        for file_name in self._datastore_path.ls():
            ## Ignore files starting with underscore. 
            ## Such files are also ignored by default by map-reduce jobs.
            ## We're also ignoring files starting with dot to ignore
            ## ".svn" directories.
            if not (fnmatch.fnmatch(file_name, "_*") or 
                    fnmatch.fnmatch(file_name, ".*")):
                path = self._datastore_path.append(file_name)
                ## We're not checking whether the path is a directory because
                ## such testing takes too much time
                #if path.is_dir():
                    #continue
                paths.append(path)
        if len(paths) == 0:
            raise error("Specified data store path is empty or is not valid")
        paths.sort()
        return paths



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import json

from avro.datafile import DataFileWriter
from avro.io import DatumWriter

from avroknife.error import error
from avroknife.utils import encapsulate_strings, dict_to_json, to_byte_string

def to_json(data_store, record_selector, printer, pretty=False):
    """Converts selected records to JSON.
    
    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object that defines which records
            should be processed.
        printer: a Printer object that is used to print the JSON.
        pretty: True if the output should be a valid, pretty-printed JSON.
    """
    first_record = True
    if pretty:
        printer.print("[", end="") 
    for record in record_selector.get_records(data_store):
        try:
            if first_record:
                first_record = False
            else:
                if pretty:
                    printer.print(",")
                else:
                    printer.print("")
            content = encapsulate_strings(record.content)
            printer.print(dict_to_json(content, pretty), end="")
        except Exception:
            error("while processing record with index {}".format(record.index))
            raise
    if pretty:
        printer.print("]")
    else:
        ## The case when no records were present
        if not first_record:
            printer.print("")

def extract(data_store, record_selector, value_field, 
    name_field=None, create_dirs=None, output_dir_path=None):
    """Extract specified field from selected records

    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object. It defines which records
            should be processed.
        value_field: name of the field to be extracted. 
            Can contain nested names, e.g. 'field1.field2' .
        name_field: value of this field is used as a name of the file 
            containing extracted data for given row. If this is not given,
            the name of the file is assumed to be the index of corresponding
            record.
        create_dirs: extracted fields corresponding to the same name 
            are to be placed in the same directory. 
            Name of directory is equal to the `name_field`; 
            names of the files inside are consecutive numbers.
        output_dir_path: a FileSystemPath object. This is where the extracted
            fields will be saved. If this is not given, the extracted fields
            are printed to stdout.
    """
    if output_dir_path is not None:
        output_dir_path.make_dirs()
    for record in record_selector.get_records(data_store):
        datum = record.content
        if datum: #check if there is any data to be written
            field_value = __get_value(datum, value_field)
            if output_dir_path is not None:
                try:
                    output_name = __get_output_name(record.index, datum, name_field)
                    output_path = output_dir_path.append(output_name)
                    file_path = None
                    if create_dirs:
                        file_path = __prepare_dir(output_path)
                    else:
                        file_path = __prepare_file(output_path, name_field) 
                    with file_path.open("w") as output:
                        output.write(to_byte_string(field_value))
                except Exception:
                    error("while processing record with index {}"\
                        .format(record.index))
                    raise
            else:
                print(to_byte_string(field_value))

def __prepare_dir(output_dir):
    if output_dir.exists():
        if not output_dir.is_dir():
            error("File with name '{}' already exists. "\
                "Unable to create a directory with the same name".\
                format(output_dir))
            raise Exception()
    else:
        output_dir.make_dirs()
    file_names = output_dir.ls()
    new_number = 0
    if len(file_names) > 0:
        file_numbers = [int(f) for f in file_names]
        new_number = max(file_numbers)+1
    new_file_name = str(new_number)
    return output_dir.append(new_file_name)

def __prepare_file(output_file, name_field):
    if output_file.exists():
        ## There is a race condition between checking 
        ## if the file exists and opening it for writing 
        ## but I don't see  an easy way to deal with it 
        ## in a way that would work for HDFS as well as 
        ## for local file system paths. Dealing with 
        ## this problem doesn't seem to be important anyway.
        error_string = "The file with name '{}' "\
            "already exists.".format(output_file) 
        if name_field is not None:
            error_string = error_string + " This is probably "\
                "because the selected values of given field "\
                "'{}' are not unique".format(name_field)
        error(error_string)
        raise Exception()
    return output_file

def __get_output_name(record_index, datum, name_field):
    file_name = str(record_index)
    if name_field is not None:
        file_name_raw = __get_value(datum, name_field)
        file_name = "null"
        if file_name_raw is not None:
            file_name_raw_str = str(file_name_raw).strip()
            if len(file_name_raw_str) > 0:
               file_name = file_name_raw_str 
    return file_name

def __get_value(datum, field_name):
    parts = field_name.split('.')
    try:
        for part in parts:
            datum = datum[part]
        return datum
    except KeyError:
        error("Field '{}' is not defined in the data".format(field_name))
        raise

def copy(data_store, record_selector, output_dir_path):
    """Dump selected records from the data store to another Avro file
    
    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object. It defines which records
            should be processed.
        output_dir_path: a FileSystemPath object. This is where the dump will
            be saved.
    """
    output_dir_path.make_dirs()
    with output_dir_path.append("content.avro").open("w") as output:
        with DataFileWriter(output, DatumWriter(), data_store.get_schema()) \
                as writer:
            records = record_selector.get_records(data_store)
            for record in records:
                writer.append(record.content)

def get_schema(data_store):
    """Get data store schema
    
    Args:
        data_store: a DataStore object
    Returns:
        JSON schema
    """
    return dict_to_json(json.loads(str(data_store.get_schema())), True)

def count(data_store, record_selector):
    """Get the number of selected records in the data store
    
    Args:
        data_store: a DataStore object
        record_selector: a RecordSelector object. It defines which records
            should be processed.
    Returns:
        number
    """
    n = 0
    for _ in record_selector.get_records(data_store):
        n = n+1
    return n



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from avroknife.test.tools import Tools

class ToolsTestCase(unittest.TestCase):
    @staticmethod
    def replace_according_to_map(key):
        my_map = {'simple': '/some/path/to/simple/file', 
            'path_with_space': '/some/path with/space',
            'name/with/slashes': '/some/other/path'}
        return my_map[key]

    def test_replace_prefix(self):
        text = 'interesting simple path:@in:simple and some other path that '\
            'follows it @in:path_with_space and some other path: '\
            '@in:name/with/slashes'
        expected = 'interesting simple path:/some/path/to/simple/file '\
            'and some other path that follows it /some/path with/space '\
            'and some other path: /some/other/path'
        actual = Tools.replace(text, '@in:', self.replace_according_to_map)
        self.assertEqual(expected, actual)



#!/usr/bin/env python

# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

import avroknife

setup(
    name=avroknife.__title__,
    version=avroknife.__version__,
    author=avroknife.__author__,
    author_email=avroknife.__email__,
    packages=find_packages(),
    include_package_data = True,
    url='https://github.com/CeON/avroknife',
    license=avroknife.__license__,
    description=avroknife.__description__,
    ## This file is generated by running appropriate target from the Makefile
    long_description=open('tmp/README.rst').read(),
    install_requires=[
        'avro>=1.7.5',
        'nose'
    ],
    ## Add ability to run tests in the code through the setup.py script
    test_suite='nose.collector',
    keywords=['Avro', 'browsing', 'manipulation'],
    classifiers=[
        'Development Status :: 5 - Production/Stable    ',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Operating System :: POSIX :: Linux',
    ],
    scripts=['scripts/avroknife']
)



# Copyright 2013-2015 University of Warsaw
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import unittest
import shutil
import tempfile
import os.path
import itertools
import filecmp
import distutils.dir_util

from avroknife.test.command_line_runner import CommandLineRunner, \
    CommandLineRunnerException
from avroknife.test import example_data_stores

class CommandLineTestCaseBase(unittest.TestCase):
    __source_data_dir = None
    _r = None
    __hdfs_tests_env_name = 'AVROKNIFE_HDFS_TESTS'
    __hdfs_tests = False
    
    @classmethod
    def setUpClass(cls):
        cls.__source_data_dir = tempfile.mkdtemp()
        distutils.dir_util.copy_tree(
            os.path.join(os.path.dirname(__file__), 'data/local_input_dir'),
            cls.__source_data_dir)
        standard_ds_path = os.path.join(cls.__source_data_dir, 'standard')
        nested_ds_path = os.path.join(cls.__source_data_dir, 'nested')
        binary_ds_path = os.path.join(cls.__source_data_dir, 'binary')
        example_data_stores.create(
            standard_ds_path, nested_ds_path, binary_ds_path)
        enforce_local = True
        env_name = CommandLineTestCaseBase.__hdfs_tests_env_name
        if os.getenv(env_name, 'FALSE')=='TRUE':
            cls.__hdfs_tests = True
            enforce_local = False
        cls._r = CommandLineRunner(
            './scripts/avroknife', cls.__source_data_dir, enforce_local)
    
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.__source_data_dir)
        cls._r.close()
        
    @staticmethod
    def _read(file_path):
        with open(file_path, 'r') as f:
            content = f.read()
        return content
    
    @staticmethod
    def __bool_to_str(x):
        if x:
            return ' '
        else:
            return 'NOT'
    
    @staticmethod
    def __get_problem_str(in_local, out_local):
        return 'Test failed when inputs are{}local and outputs are{}local.'.\
            format(CommandLineTestCaseBase.__bool_to_str(in_local), 
                   CommandLineTestCaseBase.__bool_to_str(out_local))
    
    def _iterate(self, fun):
        in_local_values = [True]
        out_local_values = [True]
        if self.__hdfs_tests:
            in_local_values = [True, False]
            out_local_values = [True, False]            
        for in_local, out_local in itertools.product(
                        in_local_values, out_local_values):
            try:
                fun(in_local, out_local)
            except Exception:
                print(self.__get_problem_str(in_local, out_local))
                raise
    
    def _check_output(self, command_args, expected_stdout, in_local, out_local):
        ret = self._r.run(command_args, in_local, out_local)
        self.assertEqual(expected_stdout, ret.get_stdout())

    def _check_output_file_raw_content(self, command_args, expected_content, 
                                    output_file_name, in_local, out_local):
        ret = self._r.run(command_args, in_local, out_local)
        actual = self._read(ret.get_output_path(output_file_name))
        self.assertEqual(expected_content, actual)

    def _check_output_avro_file(self, command_args, expected_avro_json_content, 
                     output_file_name, in_local, out_local):
        ret = self._r.run(command_args, in_local, out_local)
        output_path = ret.get_output_path(output_file_name)
        actual = self._r.run_raw('tojson local:{}'.format(output_path))
        self.assertEqual(expected_avro_json_content, actual)
 
    @staticmethod
    def _get_expected_standard_schema():
        return """\
{
    "namespace": "avroknife.test.data", 
    "type": "record", 
    "name": "User", 
    "fields": [
        {
            "type": "int", 
            "name": "position"
        }, 
        {
            "type": "string", 
            "name": "name"
        }, 
        {
            "type": [
                "int", 
                "null"
            ], 
            "name": "favorite_number"
        }, 
        {
            "type": [
                "string", 
                "null"
            ], 
            "name": "favorite_color"
        }, 
        {
            "type": [
                "bytes", 
                "null"
            ], 
            "name": "secret"
        }
    ]
}
"""

    @staticmethod
    def _get_expected_standard_contents():
        return """\
{"position": 0, "name": "Alyssa", "favorite_number": 256, "favorite_color": null, "secret": null}
{"position": 1, "name": "Ben", "favorite_number": 4, "favorite_color": "red", "secret": null}
{"position": 2, "name": "Alyssa2", "favorite_number": 512, "favorite_color": null, "secret": null}
{"position": 3, "name": "Ben2", "favorite_number": 8, "favorite_color": "blue", "secret": "MDk4NzY1NDMyMQ=="}
{"position": 4, "name": "Ben3", "favorite_number": 2, "favorite_color": "green", "secret": "MTIzNDVhYmNk"}
{"position": 5, "name": "Alyssa3", "favorite_number": 16, "favorite_color": null, "secret": null}
{"position": 6, "name": "Mallet", "favorite_number": null, "favorite_color": "blue", "secret": "YXNkZmdm"}
{"position": 7, "name": "Mikel", "favorite_number": null, "favorite_color": "", "secret": null}
"""


class GetSchemaTestsCase(CommandLineTestCaseBase):
    def test_basic(self):
        self._iterate(self.subtest_basic)
    def subtest_basic(self, in_local, out_local):
        self._check_output('getschema @in:standard', 
            self._get_expected_standard_schema(), in_local, out_local)
     
    def test_output_file(self):
        self._iterate(self.subtest_output_file)
    def subtest_output_file(self, in_local, out_local):
        self._check_output_file_raw_content('getschema @in:standard --output @out:actual.txt',
            self._get_expected_standard_schema(), 'actual.txt',
            in_local, out_local)


class ToJSONTestsCase(CommandLineTestCaseBase):     
    def test_basic(self):
        self._iterate(self.subtest_basic)
    def subtest_basic(self, in_local, out_local):
        self._check_output('tojson @in:standard', 
                    self._get_expected_standard_contents(), in_local, out_local)
   
    def test_pretty(self):
        self._iterate(self.subtest_pretty)
    def subtest_pretty(self, in_local, out_local):
        self._check_output('tojson --pretty @in:standard', """\
[{
    "position": 0, 
    "name": "Alyssa", 
    "favorite_number": 256, 
    "favorite_color": null, 
    "secret": null
},
{
    "position": 1, 
    "name": "Ben", 
    "favorite_number": 4, 
    "favorite_color": "red", 
    "secret": null
},
{
    "position": 2, 
    "name": "Alyssa2", 
    "favorite_number": 512, 
    "favorite_color": null, 
    "secret": null
},
{
    "position": 3, 
    "name": "Ben2", 
    "favorite_number": 8, 
    "favorite_color": "blue", 
    "secret": "MDk4NzY1NDMyMQ=="
},
{
    "position": 4, 
    "name": "Ben3", 
    "favorite_number": 2, 
    "favorite_color": "green", 
    "secret": "MTIzNDVhYmNk"
},
{
    "position": 5, 
    "name": "Alyssa3", 
    "favorite_number": 16, 
    "favorite_color": null, 
    "secret": null
},
{
    "position": 6, 
    "name": "Mallet", 
    "favorite_number": null, 
    "favorite_color": "blue", 
    "secret": "YXNkZmdm"
},
{
    "position": 7, 
    "name": "Mikel", 
    "favorite_number": null, 
    "favorite_color": "", 
    "secret": null
}]
""", in_local, out_local)
    
    def test_range_stdout(self):
        self._iterate(self.subtest_range_stdout)
    def subtest_range_stdout(self, in_local, out_local):
        self._check_output('tojson @in:standard --index 3-4', """\
{"position": 3, "name": "Ben2", "favorite_number": 8, "favorite_color": "blue", "secret": "MDk4NzY1NDMyMQ=="}
{"position": 4, "name": "Ben3", "favorite_number": 2, "favorite_color": "green", "secret": "MTIzNDVhYmNk"}
""", in_local, out_local)
      
    def test_range_file_out(self):
        self._iterate(self.subtest_range_file_out)
    def subtest_range_file_out(self, in_local, out_local):
        self._check_output_file_raw_content('tojson @in:standard --index 3-4 --output @out:actual.txt', """\
{"position": 3, "name": "Ben2", "favorite_number": 8, "favorite_color": "blue", "secret": "MDk4NzY1NDMyMQ=="}
{"position": 4, "name": "Ben3", "favorite_number": 2, "favorite_color": "green", "secret": "MTIzNDVhYmNk"}
""", 'actual.txt', in_local, out_local)
     
    def test_range_without_beginning(self):
        self._iterate(self.subtest_range_without_beginning)
    def subtest_range_without_beginning(self, in_local, out_local):
        self._check_output('tojson @in:standard --index -4', """\
{"position": 0, "name": "Alyssa", "favorite_number": 256, "favorite_color": null, "secret": null}
{"position": 1, "name": "Ben", "favorite_number": 4, "favorite_color": "red", "secret": null}
{"position": 2, "name": "Alyssa2", "favorite_number": 512, "favorite_color": null, "secret": null}
{"position": 3, "name": "Ben2", "favorite_number": 8, "favorite_color": "blue", "secret": "MDk4NzY1NDMyMQ=="}
{"position": 4, "name": "Ben3", "favorite_number": 2, "favorite_color": "green", "secret": "MTIzNDVhYmNk"}
""", in_local, out_local)
    
    def test_range_without_end(self):
        self._iterate(self.subtest_range_without_end)
    def subtest_range_without_end(self, in_local, out_local):
        self._check_output('tojson @in:standard --index 4-', """\
{"position": 4, "name": "Ben3", "favorite_number": 2, "favorite_color": "green", "secret": "MTIzNDVhYmNk"}
{"position": 5, "name": "Alyssa3", "favorite_number": 16, "favorite_color": null, "secret": null}
{"position": 6, "name": "Mallet", "favorite_number": null, "favorite_color": "blue", "secret": "YXNkZmdm"}
{"position": 7, "name": "Mikel", "favorite_number": null, "favorite_color": "", "secret": null}
""", in_local, out_local)
     
    def test_single_index(self):
        self._iterate(self.subtest_single_index)
    def subtest_single_index(self, in_local, out_local):
        self._check_output('tojson @in:standard --index 4', """\
{"position": 4, "name": "Ben3", "favorite_number": 2, "favorite_color": "green", "secret": "MTIzNDVhYmNk"}
""", in_local, out_local)
     
    def test_edge_case_without_beginning(self):
        self._iterate(self.subtest_edge_case_without_beginning)
    def subtest_edge_case_without_beginning(self, in_local, out_local):
        self._check_output('tojson @in:standard --index -0', """\
{"position": 0, "name": "Alyssa", "favorite_number": 256, "favorite_color": null, "secret": null}
""", in_local, out_local)
   
    def test_edge_case_without_end(self):
        self._iterate(self.subtest_edge_case_without_end)
    def subtest_edge_case_without_end(self, in_local, out_local):
        self._check_output('tojson @in:standard --index 7-', """\
{"position": 7, "name": "Mikel", "favorite_number": null, "favorite_color": "", "secret": null}
""", in_local, out_local)


class CopyTestsCase(CommandLineTestCaseBase):  
    def test_basic(self):
        self._iterate(self.subtest_basic)
    def subtest_basic(self, in_local, out_local):
        self._check_output_avro_file('copy @in:standard --output @out:whole_copy', 
            self._get_expected_standard_contents(), 'whole_copy', in_local, out_local)


class ExtractTestsCase(CommandLineTestCaseBase):     
    def test_text_fields(self):
        self._iterate(self.subtest_text_fields)
    def subtest_text_fields(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --index 2-3 --value_field name --output @out:extracted_name', 
                           in_local, out_local)
        output_path = ret.get_output_path('extracted_name')
        actual_2 = self._read(os.path.join(output_path, '2'))
        actual_3 = self._read(os.path.join(output_path, '3'))
        self.assertEqual('Alyssa2', actual_2)
        self.assertEqual('Ben2', actual_3)

    def test_create_dirs(self):
        self._iterate(self.subtest_create_dirs)
    def subtest_create_dirs(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --index 2-3 --value_field name --create_dirs --output @out:extracted_name', 
                           in_local, out_local)
        output_path = ret.get_output_path('extracted_name')
        actual_2 = self._read(os.path.join(output_path, '2', '0'))
        self.assertEqual(1, len(os.listdir(os.path.join(output_path, '2'))))
        actual_3 = self._read(os.path.join(output_path, '3', '0'))
        self.assertEqual(1, len(os.listdir(os.path.join(output_path, '3'))))
        self.assertEqual('Alyssa2', actual_2)
        self.assertEqual('Ben2', actual_3)

    def test_name_field(self):
        self._iterate(self.subtest_name_field)
    def subtest_name_field(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --index 2-3 --value_field name --name_field favorite_color --output @out:extracted_name', 
            in_local, out_local)
        output_path = ret.get_output_path('extracted_name')
        null = self._read(os.path.join(output_path, 'null'))
        blue = self._read(os.path.join(output_path, 'blue'))
        self.assertEqual('Alyssa2', null)
        self.assertEqual('Ben2', blue)

    def test_empty_name_field(self):
        self._iterate(self.subtest_empty_name_field)
    def subtest_empty_name_field(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --index 7 --value_field name --name_field favorite_color --output @out:extracted_name', 
            in_local, out_local)
        output_path = ret.get_output_path('extracted_name')
        null = self._read(os.path.join(output_path, 'null'))
        self.assertEqual('Mikel', null)

    def test_create_dirs_name_field(self):
        self._iterate(self.subtest_create_dirs_name_field)
    def subtest_create_dirs_name_field(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --index 2-3 --value_field name --name_field favorite_color --create_dirs --output @out:extracted_name', 
            in_local, out_local)
        output_path = ret.get_output_path('extracted_name')
        null = self._read(os.path.join(output_path, 'null', '0'))
        self.assertEqual(1, len(os.listdir(os.path.join(output_path, 'null'))))
        blue = self._read(os.path.join(output_path, 'blue', '0'))
        self.assertEqual(1, len(os.listdir(os.path.join(output_path, 'blue'))))
        self.assertEqual('Alyssa2', null)
        self.assertEqual('Ben2', blue)

    def test_name_field_with_repeated_names(self):
        self._iterate(self.subtest_name_field_with_repeated_names)
    def subtest_name_field_with_repeated_names(self, in_local, out_local):
        with self.assertRaises(CommandLineRunnerException):
            ## We suppress stderr because the program uses it to write
            ## some information about the error related to the fact that the 
            ## files to be created have the same name and this looks bad 
            ## in the unit tests report printed on console.
            self._r.run('extract @in:standard --index 3-7 --value_field name --name_field favorite_color --output @out:extracted_name', 
                in_local, out_local, discard_stderr=True)

    def test_create_dirs_name_field_with_repeated_names(self):
        self._iterate(self.subtest_create_dirs_name_field_with_repeated_names)
    def subtest_create_dirs_name_field_with_repeated_names(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --value_field name --name_field favorite_color --create_dirs --output @out:extracted_name', 
            in_local, out_local)
        output_path = ret.get_output_path('extracted_name')
        null0 = self._read(os.path.join(output_path, 'null', '0'))
        red = self._read(os.path.join(output_path, 'red', '0'))
        null1 = self._read(os.path.join(output_path, 'null', '1'))
        blue0 = self._read(os.path.join(output_path, 'blue', '0'))
        green = self._read(os.path.join(output_path, 'green', '0'))
        null2 = self._read(os.path.join(output_path, 'null', '2'))
        blue1 = self._read(os.path.join(output_path, 'blue', '1'))
        empty = self._read(os.path.join(output_path, 'null', '3'))

        self.assertEqual(1, len(os.listdir(os.path.join(output_path, 'red'))))
        self.assertEqual(4, len(os.listdir(os.path.join(output_path, 'null'))))
        self.assertEqual(2, len(os.listdir(os.path.join(output_path, 'blue'))))
        self.assertEqual(1, len(os.listdir(os.path.join(output_path, 'green'))))

        self.assertEqual('Alyssa', null0)
        self.assertEqual('Ben', red)
        self.assertEqual('Alyssa2', null1)
        self.assertEqual('Ben2', blue0)
        self.assertEqual('Ben3', green)
        self.assertEqual('Alyssa3', null2)
        self.assertEqual('Mallet', blue1)
        self.assertEqual('Mikel', empty)

    @staticmethod
    def __are_files_identical(path0, path1):
        return filecmp.cmp(path0, path1, shallow=False)
    
    def test_binary_fields(self):
        self._iterate(self.subtest_binary_fields)
    def subtest_binary_fields(self, in_local, out_local):
        ret = self._r.run('extract @in:binary --value_field packed_files --output @out:extracted_packed_files', 
                           in_local, out_local)
        output_path = ret.get_output_path('extracted_packed_files')
        self.assertTrue(self.__are_files_identical(
            os.path.join(os.path.dirname(__file__), 'data/binary_stuff/various_stuff.tar.gz'), 
            os.path.join(output_path, '0')))
        self.assertTrue(self.__are_files_identical(
            os.path.join(os.path.dirname(__file__), 'data/binary_stuff/greetings.tar.gz'), 
            os.path.join(output_path, '1')))
    
    def test_text_field_stdout(self):
        self._iterate(self.subtest_text_field_stdout)
    def subtest_text_field_stdout(self, in_local, out_local):
        self._check_output('extract @in:standard --index 2 --value_field name', """\
Alyssa2
""", in_local, out_local)
    
    def test_nested_fields(self):
        self._iterate(self.subtest_nested_fields)
    def subtest_nested_fields(self, in_local, out_local):
        ret = self._r.run('extract @in:nested --value_field sub.level2 --output @out:nested', 
                           in_local, out_local)
        output_path = ret.get_output_path('nested')
        self.assertEqual('2', open(os.path.join(output_path, '0'), 'r').read())
        self.assertEqual('1', open(os.path.join(output_path, '1'), 'r').read())

    def test_nested_name_field(self):
        self._iterate(self.subtest_nested_name_field)
    def subtest_nested_name_field(self, in_local, out_local):
        ret = self._r.run('extract @in:nested --value_field sup --name_field sub.level2 --output @out:nested', 
                           in_local, out_local)
        output_path = ret.get_output_path('nested')
        self.assertEqual('1', open(os.path.join(output_path, '2'), 'r').read())
        self.assertEqual('2', open(os.path.join(output_path, '1'), 'r').read())

class SelectTestsCase(CommandLineTestCaseBase):     
    def test_number(self):
        self._iterate(self.subtest_number)
    def subtest_number(self, in_local, out_local):
        self._check_output('tojson @in:standard --select position=1', """\
{"position": 1, "name": "Ben", "favorite_number": 4, "favorite_color": "red", "secret": null}
""", in_local, out_local)
    
    def test_string(self):
        self._iterate(self.subtest_string)
    def subtest_string(self, in_local, out_local):
        self._check_output('tojson @in:standard --select name=Ben', """\
{"position": 1, "name": "Ben", "favorite_number": 4, "favorite_color": "red", "secret": null}
""", in_local, out_local)
    
    def test_empty_string(self):
        self._iterate(self.subtest_empty_string)
    def subtest_empty_string(self, in_local, out_local):
        self._check_output('tojson @in:standard --select favorite_color=""', """\
{"position": 7, "name": "Mikel", "favorite_number": null, "favorite_color": "", "secret": null}
""", in_local, out_local)
    
    def test_null(self):
        self._iterate(self.subtest_null)
    def subtest_null(self, in_local, out_local):
        self._check_output('tojson @in:standard --select favorite_color=null', """\
{"position": 0, "name": "Alyssa", "favorite_number": 256, "favorite_color": null, "secret": null}
{"position": 2, "name": "Alyssa2", "favorite_number": 512, "favorite_color": null, "secret": null}
{"position": 5, "name": "Alyssa3", "favorite_number": 16, "favorite_color": null, "secret": null}
""", in_local, out_local)
    
    def test_no_records(self):
        self._iterate(self.subtest_no_records)
    def subtest_no_records(self, in_local, out_local):
        self._check_output('tojson @in:standard --select name=Ben --index 2-', '', in_local, out_local)
    
    def test_no_records_pretty(self):
        self._iterate(self.subtest_no_records_pretty)
    def subtest_no_records_pretty(self, in_local, out_local):
        self._check_output('tojson --pretty @in:standard --select name=Ben --index 2-', '[]\n', in_local, out_local)
    
    def test_copy(self):
        self._iterate(self.subtest_copy)
    def subtest_copy(self, in_local, out_local):
        self._check_output_avro_file('copy @in:standard --select name=Ben --output @out:ben_copy', 
            """\
{"position": 1, "name": "Ben", "favorite_number": 4, "favorite_color": "red", "secret": null}
""", 'ben_copy', in_local, out_local)
    
    def test_extract(self):
        self._iterate(self.subtest_extract)
    def subtest_extract(self, in_local, out_local):
        ret = self._r.run('extract @in:standard --value_field name --select name=Ben --output @out:ben_name', 
                           in_local, out_local)
        output_path = ret.get_output_path('ben_name')
        actual_1 = self._read(os.path.join(output_path, '1'))
        self.assertEqual('Ben', actual_1)
   
    def test_nested_field(self):
        self._iterate(self.subtest_nested_field)
    def subtest_nested_field(self, in_local, out_local):
        self._check_output('tojson @in:nested --select sub.level2=2', """\
{"sup": 1, "sub": {"level2": 2}}
""", in_local, out_local)
 

class LimitTestsCase(CommandLineTestCaseBase):  
    def test_basic(self):
        self._iterate(self.subtest_basic)
    def subtest_basic(self, in_local, out_local):
        self._check_output('tojson @in:standard --limit 1', """\
{"position": 0, "name": "Alyssa", "favorite_number": 256, "favorite_color": null, "secret": null}
""", in_local, out_local)
 
    def test_select_no_output(self):
        self._iterate(self.subtest_select_no_output)
    def subtest_select_no_output(self, in_local, out_local):
        self._check_output('tojson @in:standard --select name=Ben --index 2-', '', in_local, out_local)
 
    def test_select(self):
        self._iterate(self.subtest_select)
    def subtest_select(self, in_local, out_local):
        self._check_output('tojson @in:standard --select favorite_color=blue', """\
{"position": 3, "name": "Ben2", "favorite_number": 8, "favorite_color": "blue", "secret": "MDk4NzY1NDMyMQ=="}
{"position": 6, "name": "Mallet", "favorite_number": null, "favorite_color": "blue", "secret": "YXNkZmdm"}
""", in_local, out_local)
 
    def test_select_with_limit(self):
        self._iterate(self.subtest_select_with_limit)
    def subtest_select_with_limit(self, in_local, out_local):
        self._check_output('tojson @in:standard --select favorite_color=blue --limit 1', """\
{"position": 3, "name": "Ben2", "favorite_number": 8, "favorite_color": "blue", "secret": "MDk4NzY1NDMyMQ=="}
""", in_local, out_local)


class SchemaProjectionTestsCase(CommandLineTestCaseBase): 
    def test_index(self):
        self._iterate(self.subtest_index)
    def subtest_index(self, in_local, out_local):
        self._check_output('tojson @in:standard --index -4 --schema @in:user_projection.avsc', """\
{"position": 0, "name": "Alyssa"}
{"position": 1, "name": "Ben"}
{"position": 2, "name": "Alyssa2"}
{"position": 3, "name": "Ben2"}
{"position": 4, "name": "Ben3"}
""", in_local, out_local)
 
    def test_copy(self):
        self._iterate(self.subtest_copy)
    def subtest_copy(self, in_local, out_local):
        self._check_output_avro_file('copy @in:standard --schema @in:user_projection.avsc --output @out:projected', 
            """\
{"position": 0, "name": "Alyssa"}
{"position": 1, "name": "Ben"}
{"position": 2, "name": "Alyssa2"}
{"position": 3, "name": "Ben2"}
{"position": 4, "name": "Ben3"}
{"position": 5, "name": "Alyssa3"}
{"position": 6, "name": "Mallet"}
{"position": 7, "name": "Mikel"}
""", 'projected', in_local, out_local)


class CountTestsCase(CommandLineTestCaseBase):     
    def test_basic(self):
        self._iterate(self.subtest_basic)
    def subtest_basic(self, in_local, out_local):
        self._check_output('count @in:standard', '8\n', in_local, out_local)
        
    def test_file_output(self):
        self._iterate(self.subtest_file_output)
    def subtest_file_output(self, in_local, out_local):
        self._check_output_file_raw_content('count @in:standard --output @out:actual.txt' , 
            '8\n', 'actual.txt', in_local, out_local)
    
    def test_select(self):
        self._iterate(self.subtest_select)
    def subtest_select(self, in_local, out_local):
        self._check_output('count @in:standard --select name=Ben', '1\n', in_local, out_local)
